﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using HEAL.Attic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.PluginInfrastructure;

namespace HeuristicLab.Analysis.FitnessLandscape {
  [Item("QualityTrailAnalyzer", "An analyzer that creates a quality trail which can be further analyzed by several analyzers.")]
  [StorableType("E02FC8FB-85AD-42AB-8E14-EA509579C8BE")]
  public class QualityTrailMultiAnalyzer : CheckedMultiOperator<IQualityTrailAnalyzer>, IAnalyzer {
    public bool EnabledByDefault {
      get { return false; }
    }

    public override bool CanChangeName {
      get { return false; }
    }

    public IValueLookupParameter<IntValue> UpdateIntervalParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["QualityTrailUpdateInterval"]; }
    }
    public ILookupParameter<IntValue> UpdateCounterParameter {
      get { return (ILookupParameter<IntValue>)Parameters["QualityTrailUpdateCounter"]; }
    }
    public ILookupParameter<DataTable> QualityTrailParameter {
      get { return (ILookupParameter<DataTable>)Parameters["Quality Trail"]; }
    }
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters["Results"]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public ILookupParameter<BoolValue> MaximizationParameter {
      get { return (ILookupParameter<BoolValue>)Parameters["Maximization"]; }
    }

    public IntValue UpdateInterval {
      get { return UpdateIntervalParameter.Value; }
      set { UpdateIntervalParameter.Value = value; }
    }

    [StorableConstructor]
    protected QualityTrailMultiAnalyzer(StorableConstructorFlag _) : base(_) { }
    protected QualityTrailMultiAnalyzer(QualityTrailMultiAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public QualityTrailMultiAnalyzer()
      : base() {
      Parameters.Add(new ValueLookupParameter<IntValue>("QualityTrailUpdateInterval", "The interval at which the contained analyzers should be applied.", new IntValue(1)));
      Parameters.Add(new LookupParameter<IntValue>("QualityTrailUpdateCounter", "The number of times the Analyzer was called since the last update."));
      Parameters.Add(new LookupParameter<DataTable>("Quality Trail", "All historical quality values throughout the algorithm run"));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Quality", "Quality values of the whole population", 1));
      Parameters.Add(new LookupParameter<ResultCollection>("Results", "The collected results"));
      Parameters.Add(new LookupParameter<BoolValue>("Maximization", "Whether the problem is a maximization problem"));

      foreach (var analyzer in ApplicationManager.Manager.GetInstances<IQualityTrailAnalyzer>()) {
        Operators.Add(analyzer, !(analyzer is FitnessCloudAnalyzer));
      }
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new QualityTrailMultiAnalyzer(this, cloner);
    }

    public override IOperation InstrumentedApply() {
      var maximization = MaximizationParameter.ActualValue.Value;
      var qualityTrail = QualityTrailParameter.ActualValue;
      if (qualityTrail == null) {
        qualityTrail = new DataTable("Quality Trail");
        qualityTrail.Rows.Add(new DataRow("Qualities"));
        QualityTrailParameter.ActualValue = qualityTrail;
      }
      var qualities = QualityParameter.ActualValue.Select(x => x.Value);
      qualityTrail.Rows["Qualities"].Values.Add(maximization ? qualities.Max() : qualities.Min());

      var interval = UpdateIntervalParameter.ActualValue.Value;
      var counter = UpdateCounterParameter.ActualValue;
      if (counter == null) {
        counter = new IntValue(0);
        UpdateCounterParameter.ActualValue = counter;
      }
      counter.Value++;

      var next = new OperationCollection();
      if (counter.Value % interval == 0) {
        counter.Value = 0;
        foreach (var item in Operators.CheckedItems) next.Add(ExecutionContext.CreateOperation(item.Value));
      }
      next.Add(base.InstrumentedApply());
      return next;
    }
  }
}
